# -*- coding: utf-8 -*-

from queue import Queue
import signal
from threading import Event

from music2storage.connection import connect_to_drive, connect_to_youtube
from music2storage.pipeline import download_from_youtube, convert_to_mp3, upload_to_drive, delete_local_file
from music2storage.signalhandler import SignalHandler
from music2storage.worker import Worker


class Music2Storage:
    """Manages workers, queues, services for music2storage."""

    def __init__(self):
        """Initializes all the queues and sets default values for services and workers."""

        self.queues = {
            'download': Queue(),
            'convert': Queue(),
            'upload': Queue(),
            'delete': Queue(),
            'done': Queue(),
        }

        self.drive_service = None
        self.youtube_service = None
        
        self.workers = []
        self.stopper = Event()
        self.signal_handler = None

    def add_to_queue(self, url):
        """
        Adds an URL to the download queue.

        :param str url: URL to the music service track
        """

        if self.drive_service is not None:
            self.queues['download'].put(url)
        else:
            print('Drive service is not initialized. URL was not added to queue.')

    def connect_drive(self):
        """Opens browser to allow access to the Google Drive service, creates a credentials file, and returns the service instance."""

        self.drive_service = connect_to_drive()

    def connect_youtube(self):
        """Opens browser to allow access to the Youtube service, creates a credentials file, and returns the service instance."""

        self.youtube_service = connect_to_youtube()

    def start_workers(self, workers_per_task):
        """
        Creates and starts the workers, as well as attaching a handler to terminate them gracefully when a SIGINT signal is received.

        :param int workers_per_task: Number of workers to create for each task in the pipeline
        """

        if not workers:
            for _ in range(workers_per_task):
                self.workers.append(Worker(self._download, self.queues['download'], self.queues['convert'], self.stopper))
                self.workers.append(Worker(self._convert, self.queues['convert'], self.queues['upload'], self.stopper))
                self.workers.append(Worker(self._upload, self.queues['upload'], self.queues['delete'], self.stopper))
                self.workers.append(Worker(self._delete, self.queues['delete'], self.queues['done'], self.stopper))

            self.signal_handler = SignalHandler(self.workers, self.stopper)
            signal.signal(signal.SIGINT, self.signal_handler)

            for worker in self.workers:
                worker.start()

    def _download(self, url):
        """
        Downloads the file associated with the given URL.

        :param str url: URL of the video/audio file
        :return str: Filename of the file in local storage
        """
        
        return download_from_youtube(url)

    def _convert(self, file_name):
        """
        Converts the file associated with the file_name passed into a MP3 file.

        :param str file_name: Filename and extension of the file to be converted
        :return str: Filename of the new file in local storage
        """

        return convert_to_mp3(file_name, self.queues['delete'])

    def _upload(self, file_name):
        """
        Uploads the file associated with the file_name passed to the drive service.

        :param str file_name: Filename of the file to be uploaded
        :return str: Original filename passed as an argument
        """

        return upload_to_drive(self.drive_service, file_name)

    def _delete(self, file_name):
        """
        Deletes the file associated with the file_name passed from local storage.

        :param str file_name: Filename of the file to be deleted
        :return str: Filename of the file that was just deleted
        """

        return delete_local_file(file_name)

    def _notify(self):
        """TODO: Notify that the MP3 has been uploaded through a messaging service."""
        pass



# -*- coding: utf-8 -*-

from apiclient.discovery import build
from httplib2 import Http
from oauth2client import file, client, tools


def connect_to_drive():
    """
    Creates connection to the Google Drive API and returns the service object to make requests.

    :return googleapiclient.discovery.Resource:
    """

    SCOPES = 'https://www.googleapis.com/auth/drive'
    store = file.Storage('drive_credentials.json')
    creds = store.get()
    if not creds or creds.invalid:
        flow = client.flow_from_clientsecrets('client_secret.json', SCOPES)
        creds = tools.run_flow(flow, store)
    service = build('drive', 'v3', http=creds.authorize(Http()))
    return service


def connect_to_youtube():
    """
    Creates connection to the Youtube API and returns the service object to make requests.

    :return googleapiclient.discovery.Resource:
    """

    SCOPES = 'https://www.googleapis.com/auth/youtube.force-ssl'
    store = file.Storage('youtube_credentials.json')
    creds = store.get()
    if not creds or creds.invalid:
        flow = client.flow_from_clientsecrets('client_secret.json', SCOPES)
        creds = tools.run_flow(flow, store)
    service = build('youtube', 'v3', http=creds.authorize(Http()))
    return service  



# -*- coding: utf-8 -*-

import queue
from threading import Thread


class Worker(Thread):
    """Worker that processes items from an input queue and puts the result into an output queue."""

    def __init__(self, func, in_queue, out_queue, stopper):
        """
        Builds a worker by taking a function that does work on items from the input queue to be put in the output queue.

        :param function func: Function that does work on items from the input queue
        :param Queue in_queue: Input queue
        :param Queue out_queue: Output queue
        :param threading.Event stopper: Event that signals that the thread should stop execution
        """
        
        super().__init__()
        self.func = func
        self.in_queue = in_queue
        self.out_queue = out_queue
        self.stopper = stopper

    def run(self):
        """
        Method that gets run when the Worker thread is started.

        When there's an item in in_queue, it takes it out, passes it to func as an argument, and puts the result in out_queue.
        """

        while not self.stopper.is_set():
            try:
                item = self.in_queue.get(timeout=5)
            except queue.Empty:
                continue

            try:
                result = self.func(item, service=None)
            except TypeError:
                continue
            else:
                self.out_queue.put(result)



# -*- coding: utf-8 -*-

import os
import subprocess
from time import time

from ffmpy import FFmpeg
from googleapiclient.http import MediaFileUpload
from pytube import YouTube
from pytube.exceptions import RegexMatchError


def download_from_youtube(url):
    """
    Downloads an MP4 or WebM file that is associated with the video at the URL passed.

    :param str url: URL of the video to be downloaded
    :return str: Filename of the file in local storage
    """

    try:
        yt = YouTube(url)
    except RegexMatchError:
        print(f"Cannot download MP3 at {url}")
    else:
        stream = yt.streams.first()
        print(f"Download for {stream.default_filename} has started")
        start_time = time()
        stream.download()
        end_time = time()
        print(f"Download for {stream.default_filename} has finished in {end_time - start_time} seconds")

        return stream.default_filename


def convert_to_mp3(file_name, delete_queue):
    """
    Converts the file associated with the file_name passed into a MP3 file.

    :param str file_name: Filename of the original file in local storage
    :param Queue delete_queue: Delete queue to add the original file to after conversion is done
    :return str: Filename of the new file in local storage
    """

    new_file_name = os.path.splitext(file_name)[0] + '.mp3'

    ff = FFmpeg(
        inputs={file_name: None},
        outputs={new_file_name: None}
    )

    print(f"Conversion for {file_name} has started")
    start_time = time()
    ff.run(stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    end_time = time()
    print(f"Conversion for {file_name} has finished in {end_time - start_time} seconds")

    delete_queue.put(file_name)
    return new_file_name


def upload_to_drive(drive_service, file_name):
    """
    Uploads the file associated with the file_name passed to Google Drive in the Music folder (creates it in root if it doesn't exist).

    :param Resource drive_service: Google Drive service instance
    :param str file_name: Filename of the file to be uploaded
    :return str: Original filename passed as an argument (in order for the worker to send it to the delete queue)
    """

    response = drive_service.files().list(q="name='Music' and mimeType='application/vnd.google-apps.folder' and trashed=false").execute()

    try:
        folder_id = response.get('files', [])[0]['id']
    except IndexError:
        print('Music folder is missing. Creating it.')
        folder_metadata = {
            'name': 'Music',
            'mimeType': 'application/vnd.google-apps.folder'
        }
        folder = drive_service.files().create(body=folder_metadata, fields='id').execute()
        folder_id = folder.get('id')

    file_metadata = {
        'name': file_name,
        'parents': [folder_id]
    }

    media = MediaFileUpload(file_name, mimetype='audio/mpeg')
    
    print(f"Upload for {file_name} has started")
    start_time = time()
    file = drive_service.files().create(body=file_metadata, media_body=media, fields='id').execute()
    end_time = time()
    print(f"Upload for {file_name} has finished in {end_time - start_time} seconds")

    return file_name


def delete_local_file(file_name):
    """
    Deletes the file associated with the file_name passed from local storage.
    
    :param str file_name: Filename of the file to be deleted
    :return str: Filename of the file that was just deleted
    """

    try:
        os.remove(file_name)
        print(f"Deletion for {file_name} has finished")
        return file_name
    except OSError:
        pass



# -*- coding: utf-8 -*-

from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='music2storage',
      version='0.1.2',
      description='Downloads music from a service and stores it in a cloud or local storage.',
      long_description='Currently the only music service is Youtube and the only storage is Google Drive. Soundcloud and Dropbox are planned to be added soon.',
      url='https://github.com/Music-Moo/music2storage',
      author='Radu Raicea',
      author_email='radu@raicea.com',
      license='GPLv3',
      packages=['music2storage'],
      install_requires=[
          'ffmpy',
          'google-api-python-client',
          'pytube'
      ],
      include_package_data=True,
      zip_safe=False)
